/*
Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.

Redistribution of the source of this package, with or without
modification, is NOT permitted.

Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.

2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the 
   distribution, a statement that the software has been modified and
   identify the person who was responsible for the modifications.  

3. Neither the name of the author nor Forward Computing and Control may be 
   used to endorse or promote products derived from this software
   without specific prior written permission.

4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment: 
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd. 
    (http://www.forward.com.au)."


THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package au.com.forward.runWST.maps;
import au.com.forward.runWST.misc.*;
import java.util.*;
import java.util.logging.*;
import au.com.forward.webstringtemplate.filters.*;

/**
 *  An attribute hashtable with pretty printing
 *
 *@author     matthew ford
 *@created    November 26, 2003
 */
public class AttributeArrayList extends ArrayList  implements IAttributeTreeNode, Cloneable  {

		/**
		 *  Gets the childCount attribute of the ListWrapper object
		 *
		 *@param  parent  Description of the Parameter
		 *@return         The childCount value
		 */
		public int getChildCount(Object parent) {
			return size();
		}


		/**
		 *  Gets the indexOfChild attribute of the ListWrapper object
		 *
		 *@param  parent  Description of the Parameter
		 *@param  child   Description of the Parameter
		 *@return         The indexOfChild value
		 */
		public int getIndexOfChild(Object parent, Object child) {
			return indexOf(child);
		}
    

		/**
		 *  Gets the child attribute of the ListWrapper object
		 *
		 *@param  parent  Description of the Parameter
		 *@param  index   Description of the Parameter
		 *@return         The child value
		 */
		public Object getChild(Object parent, int index) {
			Object child = get(index);
				return new AttributeListEntry(index, get(index));
		}



		/**
		 *  Gets the leaf attribute of the ListWrapper object
		 *
		 *@param  node  Description of the Parameter
		 *@return       The leaf value
		 */
		public boolean isLeaf(Object node) {
			return false;
		}

		
  public String toString() {
		return AttributeToString.attributeArrayListToString(this);
	}

	public Object clone() {
		AttributeArrayList list = (AttributeArrayList)super.clone();
		// now go through each element and clone it.
		for (int i = 0; i<list.size(); i++) {
			Object obj = list.get(i);
			IWSTFilter filter = null;
			while (obj instanceof IWSTFilter) {
				filter = (IWSTFilter)obj;
				obj = filter.getObject();
			}	
			// obj should be a String, AttributeArrayList or AttributeHashtable
			if (obj instanceof String) {
				// nothing to do as invariant
			} else if (obj instanceof AttributeHashtable) {
				obj = ((AttributeHashtable)obj).clone();
			} else if (obj instanceof AttributeArrayList) {
				obj = ((AttributeArrayList)obj).clone();
			} else if (obj == null) {
				throw new IllegalStateException("Expected String or AttributeArrayList or AttributeHashtable but found null object");
			} else {
				throw new IllegalStateException("Expected String or AttributeArrayList or AttributeHashtable but found '"+obj.getClass().getName()+"'");
			}				
			if (filter != null) {
				String path = filter.getPath();
				filter = filter.newInstance();
				filter.setObject(obj);
				filter.setPath(path);
				obj = filter;
			}
			list.set(i,obj);
		}	
		return list;
	}	
	
}


