/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package au.com.forward.webstringtemplate.filters;

/**
 *  The Interface for WebStringTemplate Filters.
 *
 *@author     matthew ford
 *@created    December 28, 2003
 */
public interface IWSTFilter {

	/**
	 *  Returns a new instance of this filter class, usually a shallow clone with null object and path.
	 *
	 *@return    A new instance of this filter class
	 */
	public IWSTFilter newInstance();


	/**
	 *  returns the object being filtered, null if not set.
	 *
	 *@return    the object beinig filtered.
	 */
	public Object getObject();


	/**
	 *  sets the object being filtered.
	 *
	 *@param  obj  the object to be filtered
	 *@throws IllegalStateException if the object is this filter itself or if it wraps this filter.
	 */
	public void setObject(Object obj);


	/**
	 *  returns the filtered string of this object or "" (empty string) if object null. <br>
	 *  If an error occures calling <code>obj.toString()</code>, return the error message as the string.
	 *
	 *@return    the filtered string of this object. If <code>obj.toString()</code> throws an exception, then the exception message is returned.
	 */
	public String toString();


	/**
	 *  returns the attribute name path of the object in this filter, or null if not set
	 *
	 *@return    The path value can be null
	 */
	public String getPath();


	/**
	 *  sets the attribute name path of the object in this fiter, null if not set
	 *
	 *@param  newPath                    The new path value
	 *@throws  IllegalArgumentException  if not null and not a valid attribute name
	 */
	public void setPath(String newPath);

}

