header {
/*
 Modified by Matthew Ford  21st Dec. 2003
 modifications marked with // mod mpf
 
 [The "BSD licence"]
 Copyright (c) 2003 Terence Parr
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.antlr.stringtemplate.language;
import org.antlr.stringtemplate.*;
import java.util.*;
import java.io.*;
import au.com.forward.webstringtemplate.filters.*; // mod mpf
import au.com.forward.webstringtemplate.*; // mod mpf
}

class StringTemplateLanguageEvaluator extends TreeParser;

options {
    importVocab=StringTemplateLanguageParser;
    ASTLabelType = "org.antlr.stringtemplate.language.StringTemplateAST";
}

{
    public static class NameValuePair {
        public String name;
        public Object value;
    };

    protected StringTemplate self = null;
    protected Writer out = null;
    protected ASTExpr chunk = null;

    /** Create an evaluator using attributes from self */
    public StringTemplateLanguageEvaluator(StringTemplate self, ASTExpr chunk, Writer out) {
        this.self = self;
        this.chunk = chunk;
        this.out = out;
    }

}

action
{
    Object e=null;
}
    :   e=expr {chunk.writeAttribute(self,e,out);}
    ;

expr returns [Object value=null]
{
    Object a=null, b=null, e=null;
    Map argumentContext=null;
}
    :   #(PLUS a=expr b=expr {value = chunk.add(a,b);})
    |   value=templateApplication
    |   value=attribute
    |   value=templateInclude
    |   #(VALUE e=expr)
        // convert to string (force early eval)
        {
        StringWriter buf = new StringWriter();
        chunk.writeAttribute(self,e,buf);
        value = buf.toString();
        }
    ;

templateInclude returns [Object value=null]
{
    Map argumentContext = null;
    String name = null;
    Object n=null;
}
    :   // for includes, evaluating arg list now is ok
        #(  INCLUDE
            (   t2:ID argumentContext=argList[null]
                {name=t2.getText();}
/*
            |   #( VALUE n=expr argumentContext=argList[null] )
                {name=n.toString();}
*/
            )
         )
        {value = chunk.getTemplateText(self, name, argumentContext);}
    ;

/** Apply template(s) to an attribute; can be applied to another apply
 *  result.  I am not too efficient here as I write to a StringWriter and
 *  then dump for the common <attribute:template()> case, but I'll fix later.
 */
templateApplication returns [Object value=null]
{
Object a=null;
Vector templatesToApply=new Vector();
}
    :   #(  APPLY a=expr (template[templatesToApply])+
            {value = chunk.applyListOfAlternatingTemplates(self,a,templatesToApply);}
         )
    ;

template[Vector templatesToApply]
{
Map argumentContext = null;
Object n = null;
}
    :   #(  TEMPLATE
            (   t:ID args:. // don't eval argList now; must re-eval each iteration
                {
                String templateName = t.getText();
                StringTemplateGroup group = self.getGroup();
                StringTemplate embedded = group.getEmbeddedInstanceOf(self, templateName);
                if ( embedded!=null ) {
                    embedded.setArgumentsAST(#args);
                    templatesToApply.addElement(embedded);
                }
                }

            |	anon:ANONYMOUS_TEMPLATE
                {
                StringTemplate anonymous = anon.getStringTemplate();
                //anonymous.setArgumentContext(new HashMap());
                templatesToApply.addElement(anonymous);
                }

/*
            |   #( VALUE n=expr argumentContext=argList[null] )
                {
                String templateName = n.toString();
                StringTemplateGroup group = self.getGroup();
                StringTemplate embedded = group.getEmbeddedInstanceOf(self, templateName);
                if ( embedded!=null ) {
                    embedded.setArgumentsAST(#args);
                    templatesToApply.addElement(embedded);
                }
                }
*/
            )
         )
    ;

ifCondition returns [boolean value=false]
{
    Object a=null, b=null;
}
    :   a=ifAtom {value = self.testObjectTrue(a);}  // mod mpf was a!=null;}
    |   #(NOT a=ifAtom) {value = !self.testObjectTrue(a);} // mod mpf a==null;}
	;

ifAtom returns [Object value=null]
    :   value=expr
    ;

attribute returns [Object value=null]
{
    Object obj = null;
}
    :   #( DOT obj=attribute prop:ID )
        {value = chunk.getObjectProperty(self,obj,prop.getText());}

    |   i3:ID {value=self.getAttribute(i3.getText());
		            if (value instanceof IWSTFilter) {           // mod mpf
								   IWSTFilter filter = (IWSTFilter)value;    // mod mpf
									 String path =filter.getPath();            // mod mpf
									 if ( path.equals("")) {                   // mod mpf
									 		filter.setPath(i3.getText());// mod mpf
										} else {                                 // mod mpf 
									 		filter.setPath(path+"."+((WebStringTemplate)self).getNameForPath()+"_st."+i3.getText()); // mod mpf
										}	                                       // mod mpf
								}	                                           // mod mpf
           }

    //|   i:INT {value= new WSTPlainFilter(new Integer(i.getText()));} // mod mpf

    |   s:STRING {value= new WSTPlainFilter(s.getText());}  // mod mpf
    ;

argList[Map initialContext]
    returns [Map argumentContext=null]
{
    argumentContext = initialContext;
    if ( argumentContext==null ) {
        argumentContext=new HashMap();
    }
}
    :   #( ARGS (argumentAssignment[argumentContext])* )
	;

argumentAssignment[Map argumentContext]
{
    Object e = null;
}
	:	#( ASSIGN arg:ID e=expr {if ( e!=null ) argumentContext.put(arg.getText(),e);} )
	;

